#include <stdio.h>

#include "Trie.h"

extern char char2index[];

// return the index of search item
// return -1 if nonexist
int TrieNode::Search(char ch) {
  INS_INC(3);
  if (size>ThreshHold) 
    {INS_INC(16); return pattern[char2index[ch]]; }
  int i=0;  INS_INC(1);
  INS_INC(4);
  while ( i<size) {
    INS_INC(13); if (pattern[i]==ch) break;
    i++; INS_INC(4);
    INS_INC(3);
  }
  INS_INC(3);
  if ( i<size) {
    INS_INC(2); return i;}
  else {INS_INC(2); return -1;}
}

// return the index of inserted item
int TrieNode::Insert(char ch) {
  char i;

  if (size<=ThreshHold) {
    i=0;
    while (i<size) { 
      if (pattern[i]==ch) break;
      i++;
    }
    if (i<size) return i;
  
    // Threshold
    if (space==size && space>=ThreshHold) {
      space=MaxTnSize;
      TrieNode **tn=new (TrieNode*)[space];
      memcpy(tn, next, sizeof(TrieNode*)*size);
#if 0
      delete next;
#else
      if (size*(1+sizeof(TrieNode *))+sizeof(TrieNode *) >line_size) {
         delete next;
      }
#endif
      next=tn;

      dataType *dt=new dataType[space];
      memcpy(dt, data, sizeof(dataType)*size);
      delete data;
      data=dt;

      char *str=new char[space];
      for (i=0; i<space; i++) str[i]=-1;
      for (i=0; i<size; i++) str[char2index[pattern[i]]]=i;
      delete pattern;
      pattern=str;

      pattern[char2index[ch]]=size;
      next[size]=NULL;
      data[size++]=0;

      return size-1;
    }

    // expand space when space is not enough
    if (space==size) {
      space*=2;

#if 0
      char *str=new char[space];
      TrieNode **tn=new (TrieNode*)[space];
      memcpy(str, pattern, size);
      memcpy(tn, next, sizeof(TrieNode*)*size);
      delete pattern;
      delete next;
      pattern=str;
      next=tn;
#else
      if (line_size >= space*(sizeof(char)+sizeof(TrieNode *))+sizeof(TrieNode *)) {
         char * str= (char *) malloc(space*(sizeof(char)+sizeof(TrieNode *))+sizeof(TrieNode *));
   
         TrieNode ** tn=(TrieNode **)(str+ (space/sizeof(TrieNode *)+1)*sizeof(TrieNode *));
         memcpy(str, pattern, size);
         memcpy(tn, next, sizeof(TrieNode*)*size);
         delete pattern;
         pattern=str;
         next=tn;
      }
      else {
         char *str=new char[space];
         TrieNode **tn=new (TrieNode*)[space];
         memcpy(str, pattern, size);
         memcpy(tn, next, sizeof(TrieNode*)*size);
         delete pattern;
         if (line_size < size*(sizeof(char)+sizeof(TrieNode *))+sizeof(TrieNode *))
           delete next;
         pattern=str;
         next=tn;
      }
#endif

      dataType *dt=new dataType[space];
      memcpy(dt, data, sizeof(dataType)*size);
      delete data;
      data=dt;
    }

    pattern[size]=ch;
    next[size]=NULL;
    data[size++]=0;

    return size-1;
  }

  int pos=char2index[ch];

  // it is already there 
  if (pattern[pos]>=0) return pattern[pos];

  pattern[pos]=size;
  
  // expand space when space is not enough
  if (space==size) {
    space*=2;
    TrieNode **tn=new (TrieNode*)[space];
    memcpy(tn, next, sizeof(TrieNode*)*size);
    delete next;
    next=tn;
    dataType *dt=new dataType[space];
    memcpy(dt, data, sizeof(dataType)*size);
    delete data;
    data=dt;
  }

  next[size]=NULL;
  data[size++]=0;

  return size-1;
}

Boolean Trie::Insert(char *word, dataType d) {
  TrieNode *current;
  int index;

  if (root==NULL) {
    root=new TrieNode(tnsize);
  }
  current=root;
  while (*word) {
  
    index=current->Insert(*word);

    if (word[1]) {
      if (current->next[index]==NULL) {
	current->next[index]=new TrieNode(tnsize);
      }
      current=current->next[index];
    }
    word++;
  }

  if (current->data[index]!=0)  //data2 already there
    return False;
  else {
    current->data[index]=d;
    return True;
  }
}

dataType Trie::Search(char *word) {
  TrieNode* current=root; INS_INC(2);
  int index;
  INS_INC(7);
  while (*word) {
    INS_INC(3); if (current==NULL) {INS_INC(2); return NULL;}
    index=current->Search(*word); INS_INC(8);
    INS_INC(3); if (index<0) {INS_INC(2); return NULL;}
    INS_INC(8); if (word[1]) { INS_INC(8); current=current->next[index];}
    word++; INS_INC(4);
    INS_INC(7);
  }

  INS_INC(9); return current->data[index];
}

void Trie::print_word(FILE *dumpFp) {
  fprintf(dumpFp, "%s\n", tnList[depth]->data[index[depth]]);
}

void Trie::Dump(FILE *dumpFp) {
  tnList[0]=root;
  index[0]=0;
  depth=0;

  while (1) {
    if (tnList[depth]->data[index[depth]]!=0) print_word(dumpFp);

    if (tnList[depth]->next[index[depth]]!=NULL) {
      tnList[depth+1]=tnList[depth]->next[index[depth]];
      depth++;
      index[depth]=0;
    }
    else {
      while (depth>=0 && (++(index[depth])==tnList[depth]->size))
	depth--;
      if (depth<0) break;
    }
  }
}

